# -*- coding: UTF-8 -*-
import os

import base1x

# 执行成功
errno_success = 0
# qmt错误码基数
qmt_errno_base = 1000
# miniQMT 没有找到
errno_miniqmt_not_found = qmt_errno_base + 1
# quant1x.yaml配置文件没找到
errno_config_not_exist = qmt_errno_base + 2
# 连接miniQMT失败
errno_miniqmt_connect_failed = qmt_errno_base + 3
# 非交易日
errno_not_trade_day = qmt_errno_base + 4
# 缺少账户id
errno_not_found_account_id = qmt_errno_base + 5
# 缺少订单路径
errno_not_found_order_path = qmt_errno_base + 6

kFormatFileDate = '%Y%m%d'
kFormatOnlyDate = '%Y-%m-%d'
kFormatTimestamp = '%Y-%m-%d %H:%M:%S'
kTimestamp = '%H:%M:%S'
errBadSymbol = RuntimeError("无法识别的证券代码")


def get_quant1x_config_filename() -> str:
    """
    获取quant1x.yaml文件路径
    :return:
    """
    # 默认配置文件名
    default_config_filename = 'quant1x.yaml'
    yaml_filename = os.path.join('~', 'runtime', 'etc', default_config_filename)
    user_home = base1x.homedir()
    if not os.path.isfile(yaml_filename):
        quant1x_root = os.path.join(user_home, '.quant1x')
        yaml_filename = os.path.join(quant1x_root, default_config_filename)
        yaml_filename = os.path.expanduser(yaml_filename)
    yaml_filename = os.path.expanduser(yaml_filename)
    return yaml_filename



# coding=utf-8
import os
import time

import base1x
import pandas as pd
from xtquant import xtdata
from xtquant.xttype import StockAccount

from trader1x import utils
from trader1x.config import TraderConfig
from trader1x.log4py import logger

order_buy = 1
order_sell = 2
order_junk = 3
flag_buy = "b"
flag_sell = "s"
flag_junk = "j"


class QmtContext(object):
    """
    QMT 上下文
    TODO: 按日期切换数据
    """
    current_date: str  # 当前日期
    config_filename: str  # 配置文件名
    order_path: str  # quant1x系统输出订单的路径
    account_id: str  # 账号ID
    t89k_order_file: str  # 订单文件
    t89k_flag_ready: str  # 订单就绪标志
    t89k_flag_done: str  # 订单执行完成标志
    positions_sell_done: str  # 持仓卖出状态
    qmt_order_filename: str  # qmt系统输出的订单路径
    qmt_order_done: str  # qmt系统每日订单刷新完成标志
    qmt_positions_filename: str # qmt持仓

    def __init__(self, conf: TraderConfig):
        self._config = conf
        self.current_date = time.strftime(utils.kFormatFileDate)
        self.account_id = conf.account_id
        self.order_path = conf.order_path
        self.switch_date()

    def account(self) -> StockAccount:
        return StockAccount(self.account_id)

    def sell_is_ready(self) -> bool:
        """
        卖出条件是否就绪
        :return:
        """
        return self._config.ask_time.is_trading()

    def sell_is_auto(self) -> bool:
        """
        卖出操作是否自动一刀切
        :return:
        """
        return self._config.sell_order_auto

    def head_order_is_ready(self) -> bool:
        """
        早盘(买入)订单是否准备就绪
        :return:
        """
        if os.path.isfile(self.t89k_flag_ready) and os.path.isfile(self.t89k_order_file):
            return True
        return False

    def head_order_is_auto(self) -> bool:
        """
        是否执行 早盘订单的交易
        :return:
        """
        return self._config.head_order_auto

    def tick_order_is_auto(self) -> bool:
        """
        是否执行 盘中即时订单的交易
        :return:
        """
        return self._config.tick_order_auto

    def tail_order_is_auto(self) -> bool:
        """
        是否执行 尾盘订单的交易
        :return:
        """
        return self._config.tail_order_auto

    def can_review(self) -> bool:
        """
        是佛可以复盘
        :return:
        """
        return self._config.review_time.is_trading()

    def load_head_order(self) -> pd.DataFrame:
        """
        加载早盘订单
        :return:
        """
        df = pd.read_csv(self.t89k_order_file)
        return df

    def switch_date(self):
        """
        重置属性
        :return:
        """
        logger.warning("switch_date...")
        # self.current_date = time.strftime(utils.kFormatFileDate)
        v = xtdata.get_market_last_trade_date('SH')
        local_time = time.localtime(v / 1000)
        trade_date = time.strftime(utils.kFormatFileDate, local_time)
        self.current_date = trade_date
        logger.warning("switch_date...{}", self.current_date)
        flag = 'head'
        self.t89k_flag_ready = os.path.join(self.order_path, f'{self.current_date}-{flag}.ready')
        self.t89k_flag_done = os.path.join(self.order_path, f'{self.current_date}-{flag}-{self.account_id}.done')
        self.t89k_order_file = os.path.join(self.order_path, f'{self.current_date}-{flag}.csv')
        self.positions_sell_done = os.path.join(self.order_path, f'{self.current_date}-sell-{self.account_id}.done')
        # qmt订单文件
        self.qmt_order_filename = os.path.join(self.order_path, f'{self.account_id}-orders.csv')
        self.qmt_order_done = os.path.join(self.order_path, f'{self.account_id}-orders-{self.current_date}.done')
        # qmt持仓文件
        self.qmt_positions_filename = os.path.join(self.order_path, f'{self.account_id}-positions.csv')

    def orders_has_refreshed(self) -> bool:
        """
        当日订单是否已经刷新完成
        :return:
        """
        return self.__filelock(self.qmt_order_done)

    def push_orders_refreshed(self):
        """
        标注订单刷新已完成
        :return:
        """
        self._push_local_message(self.qmt_order_done)

    def push_head_order_buy_completed(self):
        """
        买入操作完成
        :return:
        """
        self._push_local_message(self.t89k_flag_done)
        logger.info('订单买入操作完成')

    def head_order_buy_is_finished(self) -> bool:
        """
        早盘订单是否完成
        :return:
        """
        return os.path.isfile(self.t89k_flag_done)

    def push_positions_sell_completed(self):
        """
        标记卖出操作完成
        :return:
        """
        self._push_local_message(self.positions_sell_done)

    def positions_sell_finished(self):
        """
        卖出是否操作完成
        :return:
        """
        return os.path.isfile(self.positions_sell_done)

    def check_buy_order_done_status(self, code: str) -> bool:
        """
        检查买入订单执行完成状态
        :return:
        """
        flag = self.get_order_flag(code, order_buy)
        return os.path.exists(flag)

    def push_buy_order_done_status(self, code: str):
        """
        推送买入订单完成状态
        :param code:
        :return:
        """
        flag = self.get_order_flag(code, order_buy)
        self._push_local_message(flag)

    def _push_local_message(self, filename: str):
        """
        推送消息
        :param filename:
        :return:
        """
        with open(filename, 'w') as done_file:
            pass

    def get_order_flag(self, code: str, type: int) -> str:
        """
        获取订单标识
        :param self:
        :param code:
        :param type: 1-b(buy),2-s(ell),3-j(unk)
        :return:
        """
        today = time.strftime(utils.kFormatFileDate)
        if type == order_buy:
            order_flag = flag_buy
        elif type == order_sell:
            order_flag = flag_sell
        else:
            order_flag = flag_junk
        order_flag_path = self.order_path + "/var/" + today
        base1x.mkdirs(order_flag_path)
        stock_order_flag = os.path.join(order_flag_path, f'{today}-{self.account_id}-{code}-{order_flag}.done')
        return stock_order_flag

    def fix_security_code(self, symbol: str) -> str:
        """
        调整证券代码
        :param symbol:
        :return:
        """
        security_code = ''
        if len(symbol) == 6:
            flag = self.get_security_type(symbol)
            security_code = f'{symbol}.{flag}'
        elif len(symbol) == 8 and symbol[:2] in ["sh", "sz", "SH", "SZ"]:
            security_code = symbol[2:] + '.' + symbol[:2].upper()
        else:
            raise utils.errBadSymbol
        return security_code

    def get_security_type(self, symbol: str) -> str:
        """
        获取股票市场标识
        :param symbol:  代码
        :return:
        """
        if len(symbol) != 6:
            raise utils.errBadSymbol
        code_head = symbol[:2]
        if code_head in ["00", "30"]:
            return "SZ"
        if code_head in ["60", "68"]:  # 688XXX科创板
            return "SH"
        if code_head in ["510"]:
            return "SH"
        raise utils.errBadSymbol

    def __filelock(self, filename: str) -> bool:
        """
        文件锁状态
        :return:
        """
        return os.path.isfile(filename)



# -*- coding: UTF-8 -*-
import os
import sys

import yaml
from base1x import file, application
from loguru import logger as __logger

from trader1x import utils

# 获取默认的配置文件路径
config_filename = utils.get_quant1x_config_filename()
# 转换用户路径
config_filename = os.path.expanduser(config_filename)
# 检查配置文件是否存在
if not os.path.isfile(config_filename):
    __logger.error('QMT config {}: 不存在', config_filename)
    sys.exit(utils.errno_config_not_exist)

try:
    with open(config_filename, 'r', encoding='utf-8') as f:
        result = yaml.load(f, Loader=yaml.FullLoader)
        key_basedir = "basedir"
        if isinstance(result, dict) and key_basedir in result:
            quant1x_data_path = result[key_basedir]
except Exception as e:
    quant1x_data_path = file.homedir()
quant1x_data_path = os.path.expanduser(quant1x_data_path)
_, filename, _ = application()
if filename == 'pythonservice':
    filename = 'proxy'
__log_file = f"{quant1x_data_path}/logs/{filename}.log"
# print(__log_file)
__logger.add(__log_file, encoding="utf-8", rotation="00:00", retention="10 days")

logger = __logger



# -*- coding: UTF-8 -*-
import os
import sys
import warnings
from dataclasses import dataclass

import base1x
import yaml
from base1x import TradingSession

from trader1x import utils
from trader1x.log4py import logger

warnings.filterwarnings('ignore')


def list_default_factory():
    return []


@dataclass
class TradeRule:
    """
    交易规则
    """
    id: int = -1  # 策略ID, -1无效
    auto: bool = False  # 是否自动执行
    name: str = None  # 策略名称
    flag: str = ''  # 订单标识,分早盘,尾盘和盘中
    time: TradingSession = TradingSession("09:30:00~11:30:00,13:00:00~14:56:30")  # 预览-执行操作的时间段
    weight: int = 0  # 策略权重, 默认0, 由系统自动分配
    total: int = 3  # 订单总数, 默认是3
    fee_max: float = 20000.00  # 可投入资金-最大
    fee_min: float = 10000.00  # 可投入资金-最小
    sectors: list[str] = None  # 板块, 策略适用的板块列表, 默认板块为空, 即全部个股
    ignore_margin_trading: bool = True  # 剔除两融标的, 默认是剔除

    def __init__(self, params: dict = None):
        """
        初始化
        """
        if params is None:
            return
        for key, value in params.items():
            if not hasattr(self, key):
                continue
            tmp_value = getattr(self, key)
            if isinstance(tmp_value, TradingSession):
                new_value = TradingSession(value)
                setattr(self, key, new_value)
            else:
                setattr(self, key, value)

    def enable(self) -> bool:
        """
        是否有效
        :return:
        """
        return self.auto and self.id >= 0

    def buy_enable(self) -> bool:
        """
        买入有效
        :return:
        """
        return self.enable() and self.total > 0

    def sell_enable(self) -> bool:
        """
        卖出有效
        :return:
        """
        return self.enable()

    def is_cookie_cutter_for_sell(self) -> bool:
        """
        是否一刀切卖出
        :return:
        """
        return self.sell_enable() and self.total == 0


@dataclass
class TraderConfig:
    """
    配置信息
    """
    # 账号ID
    account_id: str = ''
    # 运行路径
    order_path: str = ''
    # 时间范围 - 早盘策略
    head_time: TradingSession = TradingSession("09:27:00~14:57:00")
    # 时间范围 - 尾盘策略
    tail_time: TradingSession = TradingSession("14:45:00~14:59:50")
    # 时间范围 - 盘中订单
    tick_time: TradingSession = TradingSession("09:39:00-14:57:00")
    # 时间范围 - 持仓卖出
    ask_time: TradingSession = TradingSession("09:50:00~14:59:30")
    # 时间范围 - 撤销订单
    cancel_time: TradingSession = TradingSession("09:15:00~09:19:59, 09:30:00~14:56:59")
    # 时间范围 - 盘后复盘
    review_time: TradingSession = TradingSession("00:00:00~08:30:00, 15:01:00~23:59:59")
    # 买入持仓率, 资金控制阀值
    position_ratio: float = 0.5000
    # 印花税 - 买入, 按照成交金额, 买入0.0%
    stamp_duty_rate_for_buy: float = 0.0000
    # 印花税 - 卖出, 按照成交金额, 卖出0.1%
    stamp_duty_rate_for_sell: float = 0.0010
    # 过户费 - 双向, 按照数量收取, 默认万分之六, 0.06%
    transfer_rate: float = 0.0006
    # 券商佣金 - 双向, 按成交金额计算, 默认万分之二点五, 0.025%
    commission_rate: float = 0.00025
    # 券商佣金最低, 默认5.00
    commission_min: float = 5.00
    # 保留现金
    keep_cash: float = 10000.00
    # tick订单最大金额, 默认20000.00
    tick_order_max_amount: float = 20000.00
    # tick订单最小金额, 默认10000.00
    tick_order_min_amount: float = 10000.00
    # 买入最大金额
    buy_amount_max: float = 250000.00
    # 买入最小金额
    buy_amount_min: float = 1000.00
    # 每策略最多可买股票数量, 这里默认3
    max_stock_quantity_for_strategy: int = 3
    # 自动卖出, 默认为True
    sell_order_auto: bool = False
    # 自动交易head订单, 默认为True
    head_order_auto: bool = False
    # 自动交易tick订单, 默认为True
    tick_order_auto: bool = False
    # 自动交易tail订单, 默认为True
    tail_order_auto: bool = False

    # 启动了mimiQMT机器的代理服务的监听地址, 默认为回环地址127.0.0.1, 强烈不推荐使用0.0.0.0
    proxy_address: str = base1x.lan_address()
    # 代理服务的监听端口, 默认18168
    proxy_port: int = 18168
    # 代理服务默认工作线程数, 默认为cpu核数的二分之一
    proxy_workers: int = base1x.max_procs()
    # qmt proxy地址
    proxy_url: str = ""
    # 策略集合
    strategies: list[TradeRule] = list_default_factory
    # 可撤单的交易时段
    cancel: TradingSession = TradingSession("09:15:00~09:19:59,09:25:00~11:29:59,13:00:00~14:59:59")

    def __fix_instance(self):
        """
        加载后修复
        :return:
        """
        if isinstance(self.ask_time, str):
            ts = TradingSession(self.ask_time)
            if ts.is_valid():
                self.ask_time = ts
        if isinstance(self.cancel_time, str):
            ts = TradingSession(self.cancel_time)
            if ts.is_valid():
                self.cancel_time = ts
        if isinstance(self.tick_time, str):
            ts = TradingSession(self.tick_time)
            if ts.is_valid():
                self.tick_time = ts

    def __post_init__(self):
        """
        __init__()后调用, 调整类型
        :return:
        """
        self.__fix_instance()


# 全局变量 - 配置
__global_config = TraderConfig()


def load(config_filename: str = '') -> TraderConfig:
    """
    加载配置文件
    :return:
    """
    global __global_config
    config = TraderConfig()
    config_filename = config_filename.strip()
    if len(config_filename) == 0:
        config_filename = utils.get_quant1x_config_filename()
    config_filename = os.path.expanduser(config_filename)
    logger.info(config_filename)
    if not os.path.isfile(config_filename):
        logger.error('QMT config {}: 不存在', config_filename)
        sys.exit(utils.errno_config_not_exist)
    try:
        with open(config_filename, 'r', encoding='utf-8') as f:
            result = yaml.load(f, Loader=yaml.FullLoader)
            key_trader = "trader"
            if isinstance(result, dict) and key_trader in result:
                trader = result[key_trader]
                for key, value in trader.items():
                    if not hasattr(config, key):
                        continue
                    tmp_value = getattr(config, key)
                    if isinstance(tmp_value, TradingSession):
                        new_value = TradingSession(value)
                        setattr(config, key, new_value)
                    elif tmp_value is list_default_factory:
                        if key == 'strategies':
                            tmp_list = []
                            if isinstance(value, list):
                                for d in value:
                                    rule = TradeRule(d)
                                    tmp_list.append(rule)
                            setattr(config, key, tmp_list)
                    elif key == 'account_id':
                        setattr(config, 'account_id', str(value))
                    elif key == 'top_n':
                        setattr(config, 'max_stock_quantity_for_strategy', value)
                    else:
                        setattr(config, key, value)
    except Exception as e:
        logger.error(f"发生了一个错误：{config_filename}\n错误信息：{e}")
        logger.warning('系统将使用默认配置')
        config = TraderConfig()
    # finally:
    #     logger.warning('系统将使用默认配置')
    # 检查重点配置
    if config.account_id == '':
        logger.error('配置缺少账户id')
        sys.exit(utils.errno_not_found_account_id)
    if config.order_path == '':
        logger.error('配置缺少订单路径')
        sys.exit(utils.errno_not_found_order_path)
    __global_config = config
    return config


def get() -> TraderConfig:
    global __global_config
    return __global_config


if __name__ == '__main__':
    filename = '~/workspace/quant1x.yaml'
    config = load(filename)
    print(config)



# -*- coding: UTF-8 -*-

import sys

from trader1x import utils, config, thinktrader, context, auto
from trader1x.log4py import logger

application = 'quant1x-qmt'


def main() -> int:
    """
    定时任务调用
    :return:
    """
    logger.info('{} start...', application)
    # 0. 加载配置文件
    logger.info('加载配置...')
    conf = config.load()
    logger.info('配置信息: {}', conf)
    logger.info('加载配置...OK')
    trader = thinktrader.ThinkTrader(conf)
    # 1. 连接miniQMT
    connect_result = trader.connect()
    if connect_result == 0:
        logger.info('connect miniQmt: success')
    else:
        logger.error('connect miniQmt: failed')
        return utils.errno_miniqmt_connect_failed
    logger.info('{} start...OK', application)
    # 2. 设置账号
    ctx = context.QmtContext(conf)
    trader.set_account(ctx.account_id)
    # 3. 交易流程
    auto.head_trading(ctx, trader)
    auto.positions_sell(ctx, trader)
    # 4. 关闭
    logger.info('{} stop...', application)
    trader.close()
    logger.info('{} stop...OK', application)
    logger.info('{} shutdown', application)
    return utils.errno_success


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: UTF-8 -*-
import math
import os
import time
from typing import Any

import base1x
import pandas as pd
import win32com.client
from base1x import file
from pandas import DataFrame
from xtquant import xtdata
from xtquant.xttrader import XtQuantTrader, XtQuantTraderCallback
from xtquant.xttype import *

from trader1x import utils
from trader1x.config import TraderConfig
from trader1x.context import QmtContext
from trader1x.log4py import logger


def getuser():
    """
    获取用户名
    :return:
    """
    home = file.homedir()
    _, username = os.path.split(home)
    return username


# 缓存qmt安装路径
__gjzq_qmt_exec_path = ''


def get_gjzq_qmt_exec_path() -> str:
    """
    获取QMT安装路径
    """
    global __gjzq_qmt_exec_path
    if len(__gjzq_qmt_exec_path) > 0:
        return __gjzq_qmt_exec_path
    username = getuser()
    logger.info('user={}', username)
    qmt_exec_lnk = rf'C:\Users\{username}\AppData\Roaming\Microsoft\Windows\Start Menu\Programs\国金证券QMT交易端\启动国金证券QMT交易端.lnk'
    shell = win32com.client.Dispatch("WScript.Shell")
    shortcut = shell.CreateShortCut(qmt_exec_lnk)
    target_path = str(shortcut.Targetpath)
    paths = target_path.split(r'\bin.x64')
    exec_path = os.path.expanduser(paths[0])
    # exec_path = exec_path.replace('\\', '/')
    __gjzq_qmt_exec_path = exec_path
    return exec_path


class ThinkTrader(base1x.Singleton, XtQuantTraderCallback):
    """
    迅投XtQuant-miniQMT交易
    """
    __qmt_dir: str = None
    __xt_trader = None
    __account = None
    __config = None
    __running: bool = False

    def __init__(self, conf: TraderConfig):
        """
        初始化
        :param conf:
        """
        super().__init__()
        self.__config = conf
        self.__running = True
        # miniQMT session
        self.__miniqmt_session_id = 20221128

    def close(self):
        """
        析构方法, 销毁对象
        """
        self.stop_trader()
        logger.info("thinktrader shutdown")

    def stop_trader(self):
        self.__running = False
        if self.__xt_trader is not None:
            self.__xt_trader.stop()

    def on_disconnected(self):
        """
        连接断开
        :return:
        """
        logger.error("connection lost, 交易接口断开，即将重连")
        self.stop_trader()

    def is_running(self) -> bool:
        """
        交易通道是否正常运行
        :return:
        """
        return self.__running

    def reconnect(self):
        """
        重连
        """
        if not self.__running:
            self.stop_trader()
            self.connect()

    def connect(self, qmt_dir: str = '') -> int:
        if self.__qmt_dir is None:
            qmt_dir.strip()
            if qmt_dir == '':
                qmt_dir = os.path.join(get_gjzq_qmt_exec_path(), 'userdata_mini')
                self.__qmt_dir = qmt_dir
        logger.info("miniQmt: {}", self.__qmt_dir)
        session_id = self.__miniqmt_session_id
        logger.info("session id: {}", session_id)
        self.__xt_trader = XtQuantTrader(self.__qmt_dir, session_id, self)
        # 启动交易线程
        self.__xt_trader.start()
        # 建立交易连接，返回0表示连接成功
        connect_result = self.__xt_trader.connect()
        logger.info("connect_result={}", connect_result)
        if connect_result == 0:
            self.__running = True
            logger.info('connect miniQmt: success')
            single_funds_available = self.single_available(self.__config.max_stock_quantity_for_strategy)
            logger.warning('今天, 单一策略最大可买{}个标的, 每标的可用金额{}',
                           self.__config.max_stock_quantity_for_strategy,
                           single_funds_available)
        else:
            self.__running = False
            self.__xt_trader.stop()
            self.__xt_trader = None
            logger.error('connect miniQmt: failed')
        return connect_result

    def set_account(self, account_id, account_type='STOCK'):
        self.__account = StockAccount(account_id, account_type=account_type)
        return self.__account

    @property
    def get_account(self):
        return self.__account

    @property
    def get_trader(self):
        return self.__xt_trader

    def query_asset(self) -> XtAsset:
        """
        获取资产数据
        :return:
        """
        return self.__xt_trader.query_stock_asset(self.get_account)

    def account_available(self) -> tuple[float, float]:
        """
        可用资金
        :return:
        """
        asset = self.query_asset()
        if asset is None:
            return 0.00, 0.00
        return asset.total_asset, asset.cash

    def asset_can_trade(self) -> bool:
        """
        资产是否可交易
        :return:
        """
        _, available = self.account_available()
        return available > self.__config.keep_cash

    def single_available(self, num: int) -> float:
        """
        调整单一可用资金
        :param num: 股票数量
        :return:
        """
        if num < 1:
            return 0.00
        # 1. 查询 总资产和可用
        acc_total, acc_cash = self.account_available()
        # 2. 查询持仓可卖的股票, TODO: y如果确定了可卖出的市值, 怎么保证当日必须卖出?
        positions = self.query_positions()
        can_use_amount = 0.00
        # 3. 设置一个可卖股票的市值波动范围, 这里暂定10%
        vix = 0.10
        acc_value = 0.00  # 总市值
        for position in positions:
            acc_value += position.market_value
            if position.can_use_volume < 1:
                continue
            # 计算现价
            market_price = position.market_value / position.volume
            # 累计可卖的市值: 可卖数量 * 市价
            can_use_value = position.can_use_volume * market_price
            can_use_amount += can_use_value * (1 - vix)
        acc_value = base1x.float_round(acc_value)
        if acc_total == 0:
            # qmt有一定概率出现总资金为0的情况
            acc_total = acc_value
        can_use_amount = base1x.float_round(can_use_amount)
        # 4. 确定可用资金总量: 账户可以资金 + 当日可卖出的总市值 - 预留现金
        can_use_cash = acc_cash + can_use_amount - self.__config.keep_cash
        # 5. 计算预留仓位, 给下一个交易日留position_ratio仓位
        reserve_cash = base1x.float_round(acc_total * self.__config.position_ratio)
        # 6. 计算当日可用仓位: 可用资金总量 - 预留资金总量
        available = can_use_cash - reserve_cash
        logger.warning('账户资金, 可用: {}, 市值: {}, 预留: {}, 可买: {}, 可卖: {}', acc_cash, acc_value, reserve_cash,
                       available, can_use_amount)
        # 7. 如果当日可用金额大于资金账户的可用金额, 输出风险提示
        if available > acc_cash:
            logger.warning(
                '!!! 持仓占比[{}%], 已超过可总仓位的[{}%], 必须在收盘前择机降低仓位, 以免影响下一个交易日的买入操作 !!!',
                base1x.float_round(100 * (acc_value / acc_total)),
                base1x.float_round(100 * (1 - self.__config.position_ratio)))
        # 8. 计算单一标的可用资金
        single_funds_available = base1x.float_round(available / num)
        # 9. 检查可用资金的最大值和最小值
        if single_funds_available > self.__config.buy_amount_max:
            single_funds_available = self.__config.buy_amount_max
        elif single_funds_available < self.__config.buy_amount_min:
            return 0.00
        return single_funds_available

    def available_amount(self, stock_num: int) -> float:
        """
        计算单一标的可用金额
        :param stock_num: 股票总数
        :return:
        """
        single_funds_available = self.single_available(stock_num)
        if single_funds_available <= self.__config.buy_amount_min:
            return 0.00
        return single_funds_available

    def available_amount_for_tick(self, stock_num: int) -> float:
        """
        计算单一标的可用金额
        :param stock_num: 股票总数
        :return:
        """
        single_funds_available = self.available_amount(stock_num)
        if single_funds_available <= self.__config.tick_order_min_amount:
            return 0.00
        if single_funds_available > self.__config.tick_order_max_amount:
            # 超出单个标的最大买入金额, 按照最大金额来处理
            single_funds_available = self.__config.tick_order_max_amount
        return single_funds_available

    def get_snapshot(self, security_code: str = '') -> Any:
        """
        获得快照
        :param security_code:
        :return:
        """
        tick_list = xtdata.get_full_tick([security_code])
        if len(tick_list) != 1:
            return None
        snapshot = tick_list[security_code]
        return snapshot

    def available_price(self, price: float) -> float:
        """
        计算适合的买入价格
        :param price:
        :return:
        """
        lastPrice = price
        # 价格笼子, +2%和+0.10哪个大
        buy_price = max(lastPrice * 1.02, lastPrice + 0.10)
        # 当前价格+0.05
        # buy_price = snapshot['askPrice'][0] + 0.05
        buy_price = lastPrice + 0.05
        # 最后修订价格
        buy_price = base1x.float_round(buy_price)
        return buy_price

    def calculate_buy_fee(self, price: float, volume: int) -> float:
        """
        计算买入的总费用
        :return: 股票数量
        """
        # 1. 印花税, 按照成交金额计算, 买入没有, 卖出, 0.1%
        _stamp_duty_fee = base1x.float_round(volume * price * self.__config.stamp_duty_rate_for_buy)
        # 2. 过户费, 按照股票数量, 双向, 0.06%
        _transfer_fee = base1x.float_round(volume * self.__config.transfer_rate)
        # 3. 券商佣金, 按照成交金额计算, 双向, 0.025%
        _commission_fee = base1x.float_round(volume * price * self.__config.commission_rate)
        if _commission_fee < self.__config.commission_min:
            _commission_fee = self.__config.commission_min
        # 4. 股票市值
        _stock_fee = base1x.float_round(volume * price)
        _fee = (_stamp_duty_fee + _transfer_fee + _commission_fee + _stock_fee)
        logger.debug('综合费用:{}, 委托价格={}, 数量={}, 其中印花说={}, 过户费={}, 佣金={}, 股票={}', _fee, price,
                     volume, _stamp_duty_fee, _transfer_fee, _commission_fee, _stock_fee)
        return _fee

    def calculate_stock_volumes(self, fund: float, price: float) -> int:
        """
        可以购买的股票数量(股)
        :return: 股票数量
        """
        # 1. 印花税, 按照成交金额计算, 买入没有, 卖出, 0.1%
        # stamp_duty = volume * price * stamp_duty_rate
        _stamp_duty_fee = price * self.__config.stamp_duty_rate_for_buy
        # 2. 过户费, 按照股票数量, 双向, 0.06%
        # transfer_fee = volume * transfer_rate
        _transfer_fee = self.__config.transfer_rate
        # 3. 券商佣金, 按照成交金额计算, 双向, 0.025%
        # commissions = volume * price * commission_rate
        _commission_fee = price * self.__config.commission_rate
        # 4. 股票市值
        # _stock_fee= volume * price
        _stock_fee = price
        _fee = (_stamp_duty_fee + _transfer_fee + _commission_fee + _stock_fee)
        volume = fund / _fee
        volume = math.floor(volume / 100) * 100
        # 5. 检查买入总费用, 如果大于预计金额, 则减去100股
        _fee = self.calculate_buy_fee(price, volume)
        if _fee > fund:
            volume = volume - 100
        return volume

    def head_order_can_trade(self) -> bool:
        """
        早盘订单是否可交易
        :return:
        """
        return self.__config.head_time.is_trading()

    def tick_order_can_trade(self) -> bool:
        """
        检查盘中订单是否可以交易
        :return:
        """
        return self.__config.tick_time.is_trading()

    def tick_order_is_ready(self) -> bool:
        """
        盘中订单是否就绪
        :return:
        """
        return True

    def current_date(self) -> tuple[str, str]:
        """
        今天
        :return:
        """
        today = time.strftime(utils.kFormatOnlyDate)
        v = xtdata.get_market_last_trade_date('SH')
        local_time = time.localtime(v / 1000)
        trade_date = time.strftime(utils.kFormatOnlyDate, local_time)
        return today, trade_date

    def today_is_trading_date(self) -> bool:
        """
        今天是否交易日
        :return:
        """
        (today, trade_date) = self.current_date()
        logger.info('today={}, trade_date={}', today, trade_date)
        return today == trade_date

    def order_can_cancel(self) -> bool:
        """
        委托订单可以撤销
        :return:
        """
        return self.__config.cancel_time.is_trading()

    def order(self, code: str, order_type: int, order_volume: int, price_type: int, price: float,
              strategy_name: str = '', order_remark: str = '') -> int:
        """
        下单
        :param code: 证券代码, 例如"600000.SH"
        :param order_type: 委托类型, 23:买, 24:卖
        :param order_volume: 委托数量, 股票以'股'为单位, 债券以'张'为单位
        :param price_type: 报价类型, 详见帮助手册
        :param price: 报价价格, 如果price_type为指定价, 那price为指定的价格, 否则填0
        :param strategy_name: 策略名称
        :param order_remark: 委托备注, 极简客户端的order_remark字段有长度限制，最大23个英文字符(一个中文占3个), 超出的部分会丢弃,大qmt没有长度限制
        :return: 返回下单请求序号, 成功委托后的下单请求序号为大于0的正整数, 如果为-1表示委托失败
        """
        order_id = self.__xt_trader.order_stock(self.__account, code, order_type, order_volume, price_type, price,
                                                strategy_name, order_remark)
        return order_id

    def cancel(self, order_id: int) -> int:
        """
        撤单
        :param order_id:
        :return: 返回撤单成功或者失败, 0:成功, -1:委托已完成撤单失败, -2:未找到对应委托编号撤单失败, -3:账号未登陆撤单失败
        """
        return self.__xt_trader.cancel_order_stock(self.__account, order_id)

    def buy(self, code: str, price: float, vol: int, strategy_name: str = '', order_remark: str = '') -> int:
        """
        同步下买单
        """
        order_id = self.order(code, xtconstant.STOCK_BUY, vol, xtconstant.FIX_PRICE, price, strategy_name, order_remark)
        return order_id

    def sell(self, position: XtPosition, strategy_name: str = '', order_remark: str = '') -> int:
        """
        同步下卖单
        """
        order_id = self.order(position.stock_code, xtconstant.STOCK_SELL, position.can_use_volume,
                              xtconstant.LATEST_PRICE, -1, strategy_name, order_remark)
        return order_id

    def query_positions(self) -> list[XtPosition]:
        """
        查询持仓
        :return:
        """
        positions = self.__xt_trader.query_stock_positions(self.__account)
        return positions

    def query_trades(self) -> list[XtTrade]:
        """
        查询当日成交
        TODO: 历史成交暂时不支持
        :return:
        """
        trades = self.__xt_trader.query_stock_trades(self.__account)
        return trades

    def profit_and_loss(self, ctx: QmtContext):
        """
        盈亏统计
        :return:
        """
        positions = self.query_positions()
        if len(positions) == 0:
            return
        head = positions[0]
        keys = [key for key in dir(head) if not key.startswith('__')]
        df = pd.DataFrame([[getattr(e, key) for key in keys] for e in positions], columns=keys)
        if len(df) > 0:
            df.to_csv(ctx.qmt_positions_filename, encoding='utf-8', index=False)

    def query_orders(self, cancelable_only: bool = False) -> list[XtOrder]:
        """
        查询委托
        :return:
        """
        orders = self.__xt_trader.query_stock_orders(self.__account, cancelable_only)
        return orders

    def query_order(self, order_id: int) -> XtOrder:
        """
        查询单一委托
        :return:
        """
        order = self.__xt_trader.query_stock_order(self.__account, order_id)
        return order

    def refresh_order(self) -> DataFrame | None:
        """
        刷新委托订单
        :return:
        """
        orders = self.query_orders()
        if len(orders) == 0:
            return
        head = orders[0]
        keys = [key for key in dir(head) if not key.startswith('__')]
        df = pd.DataFrame([[getattr(e, key) for key in keys] for e in orders], columns=keys)
        if len(df) > 0:
            # 修改订单字段, 将秒数改为时间戳字符串
            key_time = 'order_time'
            df[key_time] = df[key_time].apply(lambda x: time.strftime(utils.kFormatTimestamp, time.localtime(x)))
            df = self.align_fields_for_order(df)
        return df

    def align_fields_for_order(self, df: DataFrame) -> DataFrame:
        """
        对齐 订单字段
        :param df:
        :return:
        """
        # account_id: 资金账号
        # stock_code: 证券代码, 例如"600000.SH"
        # order_id: 委托编号
        # order_sysid: 柜台编号
        # order_time: 报单时间
        # order_type: 委托类型, 23:买, 24:卖
        # order_volume: 委托数量, 股票以'股'为单位, 债券以'张'为单位
        # price_type: 报价类型, 详见帮助手册
        # price: 报价价格，如果price_type为指定价, 那price为指定的价格，否则填0
        # traded_volume: 成交数量, 股票以'股'为单位, 债券以'张'为单位
        # traded_price: 成交均价
        # order_status: 委托状态
        # status_msg: 委托状态描述, 如废单原因
        # strategy_name: 策略名称
        # order_remark: 委托备注
        cols = ['order_time', 'strategy_name', 'order_remark', 'order_type', 'stock_code', 'price_type', 'price',
                'traded_price', 'order_status', 'status_msg', 'order_volume', 'traded_volume', 'account_id', 'order_id',
                'order_sysid']
        df = df[cols]
        return df

    def total_strategy_orders(self, strategy_code: int) -> int:
        """
        统计策略订单数量
        :return:
        """
        df = self.refresh_order()
        if len(df) == 0:
            return 0
        # print(df)
        condition = (df['order_remark'] == 'tick') & (df['strategy_name'] == f'S{strategy_code}')
        df = df[condition]
        return len(df)



# -*- coding: UTF-8 -*-

import os
import signal
import sys
import time

import base1x
import pandas
import pandas as pd
from base1x import market
from xtquant import xtdata

from trader1x import utils, context, thinktrader, config
from trader1x.log4py import logger

# 应用名称
application = 'quant1x-trader'


def handler_exit(signum, frame):
    """
    退出前操作
    :param signum:
    :param frame:
    :return:
    """
    global trader
    logger.info('{} shutdown...', application)
    trader.close()
    logger.info('{} shutdown...OK', application)
    sys.exit(0)


def auto_trader() -> int:
    """
    自动化交易入口
    """
    global trader
    signal.signal(signal.SIGINT, handler_exit)
    signal.signal(signal.SIGTERM, handler_exit)
    logger.info('{} start...', application)
    # 0. 加载配置文件
    logger.info('加载配置...')
    conf = config.load()
    logger.info('配置信息: {}', conf)
    logger.info('加载配置...OK')
    trader = thinktrader.ThinkTrader(conf)
    # 1. 连接miniQMT
    connect_result = trader.connect()
    if connect_result == 0:
        logger.info('connect miniQmt: success')
    else:
        logger.error('connect miniQmt: failed')
        return utils.errno_miniqmt_connect_failed
    logger.info('{} start...OK', application)
    # 2. 设置账号
    ctx = context.QmtContext(conf)
    trader.set_account(ctx.account_id)
    # # 2.1 测试 - 持仓列表
    # trader.profit_and_loss(ctx)
    # # 2.2 测试 - 刷新订单
    # trader.refresh_order()
    # # 2.3 测试 - 复盘
    # marker_review(ctx, trader)
    # # 2.4 测试 - 统计
    # stat(ctx, trader)
    # exit(0)
    # 3. 盘中交易流程
    # # 3.1 测试单一可用金额
    # stock_total = 6
    # single_funds_available = trader.available_amount(stock_total)
    # logger.info('single_funds_available= {}', single_funds_available)
    # # 3.2 测试手续费
    # single_funds_available = 10000.00
    # buy_price = 10.00
    # buy_num = trader.calculate_stock_volumes(single_funds_available, buy_price)
    # logger.info('single_funds_available={}, price={}, volume={}', single_funds_available, buy_price, buy_num)
    # exit(0)
    # 3.3 检测新增标的
    logger.info('订单路径: {}', ctx.order_path)
    date_updated = False
    while True:
        time.sleep(1)
        logger.info("检测[交易日]...")
        # 3.3.1 检测当前日期是否最后一个交易日
        (today, trade_date) = trader.current_date()
        if today != trade_date:
            logger.warning('today={}, trade_date={}, 非交易日', today, trade_date)
            date_updated = False
            continue
        elif not date_updated:
            # 如果状态还没调整
            logger.warning('today={}, trade_date={}, 当前日期为交易日', today, trade_date)
            ctx.switch_date()
            date_updated = True
            continue
        # 3.3.2 早盘策略
        head_trading(ctx, trader)
        # 3.3.3 盘中策略
        tick_trading(ctx, trader, 81)
        tick_trading(ctx, trader, 82)
        # 3.3.4 尾盘策略
        tail_trading(ctx, trader)
        # 3.3.5 盘中卖出
        positions_sell(ctx, trader)
        # 3.3.6 盘后复盘
        marker_review(ctx, trader)
        # break

    # 4. 关闭
    logger.info('{} stop...', application)
    trader.close()
    logger.info('{} stop...OK', application)
    logger.info('{} shutdown', application)
    return 0


def head_trading(ctx: context.QmtContext, trader: thinktrader.ThinkTrader):
    """
    早盘交易
    :param ctx: 上下文
    :param trader: 交易实例
    :return:
    """
    order_remark = 'head'
    logger.info("检测[交易时段 - head]...")
    # 1. 是否早盘交易时段
    if not trader.head_order_can_trade():
        logger.warning("[交易时段 - head]...skip")
        return
    # 2. 早盘订单是否自动交易
    if not ctx.head_order_is_auto():
        logger.info('早盘交易, 禁止自动交易...')
        return
    # 3. 早盘订单是否完成交易
    if ctx.head_order_buy_is_finished():
        logger.warning("[交易时段 - head]...订单已完成")
        return
    # 4. 买入订单未就绪, 退出
    if not ctx.head_order_is_ready():
        logger.warning("[交易时段 - head]...订单未就绪")
        return
    # 5. 资金是否满足交易条件
    if not trader.asset_can_trade():
        logger.warning("[交易时段 - head]...触及最低保留金额, 暂停交易")
        return
    # 6. 获取早盘订单的dataframe
    df = ctx.load_head_order()
    stock_total = len(df)
    if stock_total == 0:
        return
    logger.info('订单数量: {}', stock_total)
    # 7. 资金是否满足交易条件
    # total_asset, cash = trader.account_available()
    # 8. 遍历订单, 数据文件只需要code和open两个字段
    for idx, stock in df.iterrows():
        # 数据日期
        order_date = stock['date']
        # 证券代码
        code = stock['code']
        security_name = stock['name']
        # 检查买入成功标识
        if ctx.check_buy_order_done_status(code):
            logger.info("stock {}: buy done", code)
            continue
        # 评估可以委托买入的价格和数量
        # 查询计算单一标的可用资金
        single_funds_available = trader.available_amount(stock_total)
        if single_funds_available <= 0:
            logger.warning('!!!已满仓!!!')
            continue
        # 获取快照
        security_code = market.fix_security_code(code)
        snapshot = trader.get_snapshot(security_code)
        # 计算溢价
        last_price = snapshot['lastPrice']
        buy_price = trader.available_price(last_price)
        # 计算可买数量
        buy_num = trader.calculate_stock_volumes(single_funds_available, buy_price)
        if buy_num < 100:
            logger.warning('单一股价过高, 分仓购买力不足1手')
            stock_total = stock_total - 1
            # 设置执行下单完成状态
            ctx.push_buy_order_done_status(code)
            continue
        # 买入操作
        # strategy_code = stock['strategy_code']
        # strategy_name = stock['strategy_name']
        strategy_code = '0'
        strategy_name = '0号策略'
        logger.warning('{}: 证券名称={}, 证券代码={}, date={}, strategy_code={}, price={}, vol={}', strategy_code,
                       security_name, security_code, order_date, strategy_code, buy_price, buy_num)
        order_id = trader.buy(security_code, buy_price, buy_num, f'S{strategy_code}', order_remark)
        logger.warning('order id: {}', order_id)
        # 设置执行下单完成状态
        ctx.push_buy_order_done_status(code)
    # 设置已完成标志文件
    ctx.push_head_order_buy_completed()
    return


def tail_trading(ctx: context.QmtContext, trader: thinktrader.ThinkTrader):
    """
    尾盘交易
    :param ctx:
    :param trader:
    :return:
    """
    logger.info("检测[交易时段 - tail]...")
    return


def tick_trading(ctx: context.QmtContext, trader: thinktrader.ThinkTrader, strategy_no: int = 81):
    """
    盘中交易
    :param ctx: 上下文
    :param trader: 交易实例
    :param strategy_no: 策略编号
    :return:
    """
    order_remark = 'tick'
    logger.info("检测[交易时段 - tick]...")
    if not trader.tick_order_can_trade():
        logger.info('非盘中交易时段, waiting...')
        return
    if not ctx.tick_order_is_auto():
        logger.info('盘中订单, 禁止自动交易...')
        return
    if trader.total_strategy_orders(strategy_no) >= 6:
        logger.info('检测[交易时段 - tick]...已达到操作标的数量')
        return
    logger.warning('检测新增标的...')
    csv_stock_pool = ctx.order_path + '/stock_pool.csv'
    stat_stock_pool = ctx.order_path + '/stock_pool.stat'
    # 如果股票池状态文件不存在, 则创建
    if not os.path.isfile(stat_stock_pool):
        base1x.touch(stat_stock_pool)
    # 获取股票池csv文件的修改时间
    update_time = os.path.getmtime(csv_stock_pool)
    last_modified = os.path.getmtime(stat_stock_pool)
    if update_time < last_modified:
        logger.warning('检测新增标的...无变化')
        return
    else:
        # 同步股票池数据文件的修改时间给股票池状态文件
        os.utime(stat_stock_pool, (update_time, update_time))
    mtime = time.localtime(update_time)
    timestamp = time.strftime(utils.kFormatTimestamp, mtime)
    logger.info('{} last modify: {}', csv_stock_pool, timestamp)
    # 3.3.3 检查当日所有的订单
    df = pd.read_csv(csv_stock_pool)
    if len(df) == 0:
        return
    # 过滤条件: 当日订单, 策略编号为81且订单状态为1
    (today, trade_date) = trader.current_date()
    condition = (df['date'] == today) & (df['strategy_code'] == strategy_no) & (df['order_status'] == 1)
    tick_orders = df[condition]
    if len(tick_orders) == 0:
        return
    stock_total = len(tick_orders)
    logger.warning('盘中水位观测: {}', stock_total)
    # 遍历订单
    for idx, stock in tick_orders.iterrows():
        # print(stock)
        order_date = stock['date']
        code = stock['code']
        strategy_code = stock['strategy_code']
        strategy_name = stock['strategy_name']
        security_name = stock['name']
        security_code = market.fix_security_code(code)
        # 检查买入状态
        if ctx.check_buy_order_done_status(code):
            # 已经买入的票, 跳过
            continue
        # 评估可以委托买入的价格和数量
        # 查询计算单一标的可用资金
        single_funds_available = trader.available_amount_for_tick(stock_total)
        if single_funds_available <= 0.00:
            logger.warning('!!!已满仓!!!')
            # 设置执行下单完成状态
            ctx.push_buy_order_done_status(code)
            continue
        # 获取快照
        snapshot = trader.get_snapshot(security_code)
        # 计算溢价
        last_price = snapshot['lastPrice']
        buy_price = trader.available_price(last_price)
        # 计算可买数量
        buy_num = trader.calculate_stock_volumes(single_funds_available, buy_price)
        if buy_num < 100:
            logger.warning('单一股价过高, 分仓购买力不足1手')
            stock_total = stock_total - 1
            # 设置执行下单完成状态
            ctx.push_buy_order_done_status(code)
            continue
        logger.warning('{}: 证券名称={}, 证券代码={}, date={}, strategy_code={}, price={}, vol={}', strategy_code,
                       security_name, security_code, order_date, strategy_code, buy_price, buy_num)
        # 委托买入
        order_id = trader.buy(security_code, buy_price, buy_num, f'S{strategy_code}', order_remark)
        logger.warning('order id: {}', order_id)
        # 设置执行下单完成状态
        ctx.push_buy_order_done_status(code)


def positions_sell(ctx: context.QmtContext, trader: thinktrader.ThinkTrader):
    """
    持仓卖出
    :param ctx:
    :param trader:
    :return:
    """
    # 1. 是否自动卖出
    if not ctx.sell_is_auto():
        logger.warning('sell: 非自动卖出, 需要人工干预')
        return
    # 2. 是否卖出时段
    if not ctx.sell_is_ready():
        logger.warning('sell: 非持仓卖出时段')
        return
    # 3. 检查是否已完成卖出操作
    if ctx.positions_sell_finished():
        logger.warning('sell: 持仓卖出操作已完成')
        return
    # 4. 查询当日所有的持仓
    positions = trader.query_positions()
    logger.info("sell: 当前持仓: {}", len(positions))
    for position in positions:
        if position.can_use_volume < 100:
            continue
        # print(repr(dt))
        security_code = position.stock_code
        snapshot = trader.get_snapshot(security_code)
        # 获取股票涨停价
        stock_detail = xtdata.get_instrument_detail(security_code)
        up_stop_price = stock_detail['UpStopPrice']
        last_price = snapshot['lastPrice']
        op_flag = 'Unknown'
        if last_price < up_stop_price:
            try:
                # 卖出操作
                order_id = trader.sell(position, 't89k_sell', '测试卖出')
            except Exception:
                order_id = -1
            print('order_id:', order_id)
            op_flag = 'ASKING'
        else:
            op_flag = 'WAITING - LimitUp'
        # 控制台输出持仓记录
        logger.warning("ask: code=%s, last_price=%f, up_stop_price=%f", security_code, last_price, up_stop_price)
        logger.warning(
            "sell: %s stock %s holding: %d can_ask: %d, open_price:%f, market_value: %f" % (op_flag, security_code,
                                                                                            position.volume,
                                                                                            position.can_use_volume,
                                                                                            position.open_price,
                                                                                            position.market_value))
    # 输出卖出完成操作标识
    ctx.push_positions_sell_completed()


def marker_review(ctx: context.QmtContext, trader: thinktrader.ThinkTrader):
    """
    盘后复盘
    :return:
    """
    if not ctx.can_review():
        return
    if ctx.orders_has_refreshed():
        logger.warning('复盘操作, 已完成')
        return
    # 盈亏统计
    trader.profit_and_loss(ctx)
    # 更新订单
    order_filename = ctx.qmt_order_filename
    df = trader.refresh_order()
    if df is not None:
        # old = pd.read_csv(order_filename, encoding='utf-8', index_col=0)
        old = pd.read_csv(order_filename, encoding='utf-8')
        if len(old) > 0:
            _, trade_date = trader.current_date()
            old.drop(old[old.order_time >= trade_date].index, inplace=True)
            old = trader.align_fields_for_order(old)
            df = pandas.concat([old, df])
        df.sort_values(by='order_id', inplace=True)
        df.to_csv(order_filename, encoding='utf-8', index=False)
        ctx.push_orders_refreshed()


def stat(ctx: context.QmtContext, trader: thinktrader.ThinkTrader):
    """
    统计
    :param ctx:
    :param trader:
    :return:
    """
    csv_stock_pool = ctx.order_path + '/stock_pool.csv'
    df = pd.read_csv(csv_stock_pool)
    if len(df) == 0:
        return
    # 过滤条件: 当日订单, 策略编号为81且订单状态为1
    (today, trade_date) = trader.current_date()
    # condition = (df['date'] == today) & (df['strategy_code'] == 81)
    condition = (df['date'] == trade_date) & (df['strategy_code'] == 81)
    tick_orders = df[condition].copy()
    tick_orders['sell'] = 0.00
    if len(tick_orders) == 0:
        return
    stock_total = len(tick_orders)
    print(stock_total)
    for idx, stock in tick_orders.iterrows():
        code = stock['code']
        security_code = market.fix_security_code(code)
        snapshot = trader.get_snapshot(security_code)
        last_price = snapshot['lastPrice']
        # stock['sell'] = last_price
        tick_orders.at[idx, 'sell'] = float(last_price)
    tick_orders['x'] = 100 * (tick_orders['sell'] - tick_orders['buy']) / tick_orders['buy']
    df = tick_orders[['code', 'name', 'buy', 'sell', 'x', 'speed', 'active']]
    print(df)
    df1 = df[df['x'] > 0]
    print(df1)
    print(df['x'].sum() / len(df))


if __name__ == '__main__':
    sys.exit(auto_trader())



# -*- coding: UTF-8 -*-
import signal

import base1x
import xtquant.xtdata
from base1x import market

module = 'trader1x'

base1x.redirect(module, __file__)

import sys
from path import Path
import uvicorn
from fastapi import FastAPI, Form
from xtquant import xtconstant

from trader1x import thinktrader, config, utils, context
from trader1x.log4py import logger, quant1x_data_path

__application_proxy = 'quant1x-proxy'

order_status = {
    xtconstant.ORDER_UNREPORTED: '未报',
    xtconstant.ORDER_WAIT_REPORTING: '待报',
    xtconstant.ORDER_REPORTED: '已报',
    xtconstant.ORDER_REPORTED_CANCEL: '已报待撤',
    xtconstant.ORDER_PARTSUCC_CANCEL: '部成待撤',
    xtconstant.ORDER_PART_CANCEL: '部撤',
    xtconstant.ORDER_CANCELED: '已撤',
    xtconstant.ORDER_PART_SUCC: '部成',
    xtconstant.ORDER_SUCCEEDED: '已成',
    xtconstant.ORDER_JUNK: '废单',
    xtconstant.ORDER_UNKNOWN: '未知'
}

# 操作 - 状态码 - 成功
OPERATION_STATUS_SUCCESS = 0
# 操作 - 信息 - 成功
OPERATION_MESSAGE_SUCCESS = 'success'
# 操作 - 状态码 - 未知
OPERATION_STATUS_UNKNOWN = 999
# 操作 - 信息 - 未知
OPERATION_MESSAGE_UNKNOWN = 'unknown'

__uri_prefix = '/qmt'
app = FastAPI(root_path=__uri_prefix)


def trader_reconnect():
    """
    交易通道重连
    """
    global __context, __trader, __config
    ctx_need_init = __trader.is_running()
    __trader.reconnect()
    if not ctx_need_init and __trader.is_running():
        __context = context.QmtContext(__config)


from apscheduler.schedulers.background import BackgroundScheduler

# 1.1
__scheduler = BackgroundScheduler(timezone="Asia/Shanghai")
__scheduler.add_job(trader_reconnect, trigger='interval', seconds=10)
__scheduler.start()


@app.on_event('startup')
async def proxy_init():
    """
    代理初始化
    :return:
    """
    # global __config
    global __context, __trader, __config
    logger.info('{} start...', __application_proxy)
    # 1. 获取配置信息
    __config = config.load()
    __trader = thinktrader.ThinkTrader(__config)
    # 2. 设置账号
    # __context = context.QmtContext(__config)
    __trader.set_account(__config.account_id)

    # # 1.1
    # __scheduler = BackgroundScheduler(timezone="Asia/Shanghai")
    # __scheduler.add_job(trader_reconnect, trigger='interval', second=10)
    # __scheduler.start()
    # 3. 连接miniQMT
    connect_result = __trader.connect()
    if connect_result == 0:
        logger.info('connect miniQmt: success')
    else:
        logger.error('connect miniQmt: failed')
        return utils.errno_miniqmt_connect_failed
    logger.info('{} start...OK', __application_proxy)
    # 4. 设置账号
    __context = context.QmtContext(__config)
    # __trader.set_account(__context.account_id)
    # single_funds_available = __trader.single_available(__config.max_stock_quantity_for_strategy)
    # logger.warning('今天, 单一策略最大可买{}个标的, 每标的可用金额{}', __config.max_stock_quantity_for_strategy,
    #                single_funds_available)


@app.on_event('shutdown')
async def proxy_shutdown():
    """
    代理关闭
    :return:
    """
    logger.info('{} shutdown...', __application_proxy)
    __trader.close()
    logger.info('{} shutdown...OK', __application_proxy)


@app.api_route('/health')
async def hello(name: str = 'anonymous'):
    """
    探测服务接口(√)
    :param name:
    :return:
    """
    return {rf'hello, {name}'}


@app.api_route('/data/subscribe_whole_quote')
async def data_subscribe_whole_quote(code: str = 'SH,SZ'):
    """
    订阅全推行情
    :param code:
    :return:
    """
    code_list = code.split(",")
    data = xtquant.xtdata.subscribe_whole_quote(code_list)
    return {'status': OPERATION_STATUS_SUCCESS, 'message': OPERATION_MESSAGE_SUCCESS, 'seq': data}


@app.api_route('/data/unsubscribe_quote')
async def data_unsubscribe_quote(seq: str = ''):
    """
    反订阅行情数据
    :param seq: 订阅号
    :return:
    """
    seq_id = int(seq)
    data = xtquant.xtdata.unsubscribe_quote(seq_id)
    return data


@app.api_route('/data/tick')
async def data_full_tick(code: str = 'SH,SZ'):
    """
    获取tick(√)
    :param code:
    :return:
    """
    code_list = code.split(",")
    data = xtquant.xtdata.get_full_tick(code_list)
    return data


@app.api_route('/query/asset', methods=['GET', 'POST'])
async def query_assets():
    """
    查询总资产(√)
    """
    asset = __trader.query_asset()
    return {"total_asset": asset.total_asset,
            "cash": asset.cash,
            "market_value": asset.market_value,
            "frozen_cash": asset.frozen_cash
            }


@app.api_route('/query/holding', methods=['GET', 'POST'])
async def query_holding():
    """
    查询当前持仓(√)
    """
    holding = []
    for p in __trader.query_positions():
        holding.append(
            {'account_type': p.account_type,
             'account_id': p.account_id,
             'stock_code': p.stock_code,
             'volume': p.volume,
             'can_use_volume': p.can_use_volume,
             'open_price': base1x.fix_float(p.open_price),
             'market_value': base1x.fix_float(p.market_value),
             'frozen_volume': p.frozen_volume,
             'on_road_volume': p.on_road_volume,
             'yesterday_volume': p.yesterday_volume,
             'avg_price': base1x.fix_float(p.avg_price),
             }
        )
    return holding


@app.api_route('/query/trade', methods=['GET', 'POST'])
async def query_trade():
    """
    查询当日成交(√)
    """
    trades = __trader.query_trades()
    result = []
    for trade in trades:
        result.append(
            {'stock_code': trade.stock_code,
             'order_type': trade.order_type,
             'traded_volume': trade.traded_volume,
             'traded_price': trade.traded_price,
             'traded_amount': trade.traded_amount,
             'traded_time': base1x.seconds_to_timestamp(trade.traded_time),
             "traded_id": trade.traded_id, "order_sysid": trade.order_sysid})
    return result


@app.api_route('/query/order', methods=['GET', 'POST'])
async def query_order(order_id: str = ''):
    """
    查询当日委托(√)
    :param order_id: 订单id
    """
    order_id = order_id.strip()
    if order_id == '' or order_id == '0':
        orders = __trader.query_orders()
    else:
        order = __trader.query_order(int(order_id))
        # 订单不存在，下单失败
        if order is None:
            return []
        orders = [order]
    result = []
    for order in orders:
        result.append(
            {'account_type': order.account_type,
             'account_id': order.account_id,
             'order_time': base1x.seconds_to_timestamp(order.order_time),
             'stock_code': order.stock_code,
             'order_type': order.order_type,
             'price': order.price,
             'price_type': order.price_type,
             'order_volume': order.order_volume,
             'order_id': order.order_id,
             "order_sysid": order.order_sysid,
             'traded_price': order.traded_price,
             'traded_volume': order.traded_volume,
             'order_status': order.order_status,
             'status_msg': order.status_msg,
             'strategy_name': order.strategy_name,
             'order_remark': order.order_remark,
             }
        )
    return result


def throw_error(status: int, message: str) -> dict:
    """
    抛出错误
    :param status: 状态码, 0-成功, 非0-失败
    :param message: 错误信息
    :return:
    """
    return {'status': status, 'message': message}


@app.api_route('/trade/order', methods=['POST'])
async def trade_place_order(direction: str = Form(''),
                            code: str = Form(''),
                            price_type: str = Form(''),
                            price: str = Form(''),
                            volume: str = Form(''),
                            strategy: str = Form(''),
                            remark: str = Form('')
                            ):
    """
    下单(√), 参数有:
    direction: 交易方向, buy, sell
    code: 证券代码, 格式:{code}.{marker_id}
    price_type: 报价类型
    price: 价格, 单位是元
    volume: 数量, 单位是股
    strategy: 策略名称
    remark: 订单备注
    """
    order_errno = 10000
    status = OPERATION_STATUS_SUCCESS
    message = OPERATION_MESSAGE_SUCCESS
    # 1. 交易方向
    direction = direction.strip()
    if direction == 'buy':
        # 1.1 买入
        trade_direction = xtconstant.STOCK_BUY
    elif direction == 'sell':
        # 1.2 卖出
        trade_direction = xtconstant.STOCK_SELL
    else:
        # 1.3 方向错误
        return throw_error(order_errno + 10, '交易方向错误')
    # 2. 证券代码
    code = code.strip()
    if code == '':
        return throw_error(order_errno + 20, '证券代码不能为空')
    elif len(code) != 9:
        return throw_error(order_errno + 21, '非A股证券代码长度')
    elif not code.endswith(market.tup_market):
        return throw_error(order_errno + 22, '交易所简写错误')
    stock_code = code
    # 3. 报价类型
    price_type = price_type.strip()
    if price_type == '':
        return throw_error(order_errno + 30, '报价类型不能为空')
    stock_price_type = int(price_type)
    if stock_price_type != xtconstant.FIX_PRICE and stock_price_type != xtconstant.LATEST_PRICE:
        return throw_error(order_errno + 30, '报价类型必须是最新价或限价')
    # 4. 价格
    price = price.strip()
    if price == '':
        return throw_error(order_errno + 40, '委托价格不能为空')
    stock_price = float(price)
    if stock_price <= 1.000:
        return throw_error(order_errno + 41, '委托价格不能出现小于等于1.000元')
    # 5. 数量
    volume = volume.strip()
    if volume == '':
        return throw_error(order_errno + 50, '交易数量不能为空')
    stock_volume = int(volume)
    if stock_volume % 100 != 0:
        return throw_error(order_errno + 51, '交易数量非100股的整数倍')
    # 6. 策略名称
    strategy = strategy.strip()
    if strategy == '':
        return throw_error(order_errno + 60, '策略名称不能为空')
    strategy_name = strategy
    # 7. 订单备注
    remark = remark.strip()
    if remark == '':
        return throw_error(order_errno + 70, '订单备注不能为空')
    if len(remark.encode('utf-8')) > 24:
        return throw_error(order_errno + 71, '订单备注不能超过24个字节')
    order_remark = remark
    # 8. 执行同步委托下单
    order_id = __trader.order(stock_code, trade_direction, stock_volume, stock_price_type, stock_price,
                              strategy_name, order_remark)
    logger.warning('order[{}]: code={}, direction={}, price={}, volume={}, strategy_name={}, order_remark={}', order_id,
                   stock_code, direction, stock_price, stock_volume, strategy_name, order_remark)
    return {'status': status, 'message': message, 'order_id': order_id}


@app.api_route('/trade/cancel', methods=['POST'])
async def trade_cancel_order(order_id: str = Form('')):
    """
    撤单(√)
    """
    cancel_errno = 20000
    order_id = order_id.strip()
    if order_id == '':
        return throw_error(cancel_errno + 1, 'order_id不能为空')
    elif not order_id.isdigit():
        return throw_error(cancel_errno + 2, 'order_id必须是整型')
    cancel_order_id = int(order_id)
    if cancel_order_id <= 0:
        return throw_error(cancel_errno + 3, 'order_id必须大于0')
    result = __trader.cancel(cancel_order_id)
    logger.warning(f'order_id={order_id}, errno={result}')
    # 返回撤单成功或者失败, 0: 成功, -1: 委托已完成, -2: 未找到对应委托编号, -3: 账号未登陆
    if result == 0:
        return throw_error(OPERATION_STATUS_SUCCESS, OPERATION_MESSAGE_SUCCESS)
    elif result == -1:
        return throw_error(cancel_errno + 4, '委托已完成')
    elif result == -2:
        return throw_error(cancel_errno + 5, '未找到对应委托编号')
    elif result == -3:
        return throw_error(cancel_errno + 6, '账号未登陆')
    else:
        logger.warning(f'cancel: order_id={order_id}, 未知错误, errno={result}')
    return throw_error(cancel_errno + OPERATION_STATUS_UNKNOWN, OPERATION_MESSAGE_UNKNOWN)


def sign(enable: bool = True):
    """
    验签 - 函数装饰器(aspect)
    TODO: 具体验签方式未完成
    :param enable: 是否启用验签, 默认启用验签
    :return:
    """

    def count_time(func):
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        return wrapper

    return count_time


def proxy_exit(signum, frame):
    """
    退出前操作
    :param signum:
    :param frame:
    :return:
    """
    global trader
    logger.info('{} shutdown...', __application_proxy)
    proxy_stop()
    logger.info('{} shutdown...OK', __application_proxy)
    sys.exit(0)


def proxy_start() -> int:
    """
    启动代理
    :return:
    """
    global web_server
    # signal.signal(signal.SIGINT, proxy_exit)
    # signal.signal(signal.SIGTERM, proxy_exit)
    # 1. 加载配置文件
    # logger.add('d:\proxy.log', rotation='00:00')
    logger.info('加载配置...')
    __config = config.load()
    logger.info('配置信息: {}', __config)
    logger.info('加载配置...OK')
    # 2. 配置路由
    # app.include_router(prefix=__uri_prefix)
    # 3. 启动服务
    logger.warning('{} start http server[{}:{}]...', __application_proxy, __config.proxy_address, __config.proxy_port)
    # uvicorn.run(app=f'{module}.{Path(__file__).stem}:app', host=__config.proxy_address, port=__config.proxy_port,
    #             workers=__config.proxy_workers,
    #             reload=False, debug=True)
    # uvicorn.run(app='trader1x.proxy:app', host=__config.proxy_address, port=__config.proxy_port,
    #             workers=__config.proxy_workers,
    #             reload=False, debug=True)

    log_config = {
        "version": 1,
        "disable_existing_loggers": True,
        "handlers": {
            "file_handler": {
                "class": "logging.FileHandler",
                "filename": f'{quant1x_data_path}/logs/logfile.log',
            },
        },
        "root": {
            "handlers": ["file_handler"],
            "level": "INFO",
        },
    }
    cfg = uvicorn.Config(app=f'{module}.{Path(__file__).stem}:app', host=__config.proxy_address,
                         port=__config.proxy_port,
                         workers=__config.proxy_workers, log_config=log_config)
    web_server = uvicorn.Server(config=cfg)
    # According to https://github.com/encode/uvicorn/issues/526, on Python 3.8+ you MIGHT also need:
    # asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    # But in my case it didn't seem to be necessary
    web_server.install_signal_handlers = lambda: None  # this is the necessary workaround
    web_server.run()
    return 0


def proxy_stop():
    """
    关闭代理
    :return:
    """
    global web_server
    # web_server.shutdown()
    try:
        # web_server.shutdown()
        web_server.should_exit = True
        web_server.shutdown()
    except Exception as e:
        logger.exception(e)


import socket
import servicemanager
import win32event
import win32service
from trader1x import win32serviceutil

cmd_service = 'service'


class ProxyService(win32serviceutil.ServiceFramework):
    _svc_name_ = 'quant1x-qmt-proxy'
    _svc_display_name_ = 'Quant1X-miniQMT-Proxy'
    _svc_description_ = 'Quant1X miniQMT 代理服务'
    _exe_args_ = cmd_service

    def __init__(self, args):
        """
        Constructor of the winservice
        """
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        socket.setdefaulttimeout(60)
        self.is_alive = True

    def SvcStop(self):
        """
        Called when the service is asked to stop
        """
        self.stop()
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        """
        Called when the service is asked to start
        """
        self.start()
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                              servicemanager.PYS_SERVICE_STARTED,
                              (self._svc_name_, ''))
        self.main()

    def start(self):
        """
        Override to add logic before the start
        eg. running condition
        """
        self.is_alive = True

    def stop(self):
        """
        Override to add logic before the stop
        eg. invalidating running condition
        """
        self.is_alive = False
        proxy_stop()

    def main(self):
        """
        Main class to be overridden to add logic
        """
        try:
            proxy_start()
        except Exception as e:
            logger.exception(e)


def main() -> int:
    argv = sys.argv
    argc = len(argv)
    if argc <= 1:
        # 直接启动proxy
        signal.signal(signal.SIGINT, proxy_exit)
        signal.signal(signal.SIGTERM, proxy_exit)
        ret = proxy_start()
    elif argv[1] == cmd_service:
        argv.remove(cmd_service)
        if len(argv) == 1:
            servicemanager.Initialize()
            servicemanager.PrepareToHostSingle(ProxyService)
            servicemanager.StartServiceCtrlDispatcher()
            ret = 0
        else:
            ret = win32serviceutil.HandleCommandLine(ProxyService, argv=argv)
    else:
        print('usage: quant1x-qmt-proxy')
        print('       quant1x-qmt-proxy service [install|remove|start|stop|restart]')
        ret = 0
    return ret


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# coding=utf-8
"""The setup script."""
import setuptools
from base1x.version import project_version

from trader1x import __author__

try:
    from setuptools import find_packages, setup
except ImportError:
    from distutils.core import find_packages, setup


latest, author = project_version()
__version__ = latest

def parse_requirements(filename):
    line_iter = (line.strip() for line in open(filename))
    return [line for line in line_iter if line and not line.startswith("#")]


with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.md", encoding="utf-8") as history_file:
    history = history_file.read()

requirements = parse_requirements("requirements.txt")
test_requirements = requirements

setuptools.setup(
    name="quant1x-trader",
    description="Quant1X程序化自动化交易",
    author_email="wangfengxy@sina.cn",
    url="https://gitee.com/quant1x/trader",
    version=__version__,
    author=__author__,
    long_description=readme,
    packages=setuptools.find_packages(include=["trader1x", "trader1x.*"]),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="quant1x auto trader",
    entry_points={
        "console_scripts": [
            "quant1x-auto-trader=trader1x.auto:auto_trader",
            "quant1x-qmt=trader1x.qmt:main",
            "quant1x-qmt-proxy=trader1x.proxy:main",
        ]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    data_files=[
        # ('xtquant', ['xtquant/xtdata.ini', 'xtquant/xtdata.log4cxx']),
    ],
    package_data={
        '': ['*.dll', '*.pyd', '*.ini', '*.log4cxx'],
    },
    test_suite="tests",
    tests_require=test_requirements,
    setup_requires=requirements,
)



import time

import pandas as pd
import xtquant.xtdata

# 禁止显示XtQuant的hello信息
xtquant.xtdata.enable_hello = False
code = "SH,SZ"
code_list = code.split(",")
start = time.time()
data = xtquant.xtdata.get_full_tick(code_list)
end = time.time()
print(f"耗时: {end - start:.4f} 秒")
df = pd.DataFrame(data)
print(df['600600.SH'])
